/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006-2007   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * file.c - file access functions
 */

#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_SYS_FILE_H
	#include <sys/file.h>
#endif /* HAVE_SYS_FILE_H */

#ifdef HAVE_SYS_PARAM_H
	#include <sys/param.h>
#endif /* HAVE_SYS_PARAM_H */

#ifdef HAVE_SYS_STAT_H
	#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */

#ifdef HAVE_SYS_TYPES_H
	#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */

#include <ctype.h>
#include <curses.h>

#ifdef HAVE_DIRENT_H
	#include <dirent.h>
#endif /* HAVE_DIRENT_H */

#include <errno.h>

#ifdef HAVE_FCNTL_H
	#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#include <stdio.h>

#ifdef HAVE_STDLIB_H
	#include <stdlib.h>
#endif /* HAVE_STDLIB_H */

#ifdef HAVE_STRING_H
	#include <string.h>
#endif /* HAVE_STRING_H */

#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include "gettext.h"
#include "pathnames.h"
#include "typespeed.h"

#define _(string)	gettext(string)
#define WORDFILE_MAX	100

struct hentry {
	unsigned long count;
	unsigned long tcount;
	unsigned long wordswritten;
	char name[41];
	char mod[NAME_MAX * 2 + 1];
	char wordlist[NAME_MAX * 2 + 1];
	clock_t duration;
	unsigned int sinit;
};

extern unsigned long	 cstrl(char *);
extern void		 defrule(void);
extern void		 drawscreen(void);
extern int		 fileselmenu(int, struct finfo *, const char *);
extern unsigned short	 level(int);
extern void		 liima_mvgetnstr(int, int, char *, int);
extern void		 nowordlist(void);
extern void		 pressanykey(int, int);
extern int		 typorankkaus(float);
extern int		 wanttoreset(void);
extern void		 xcolor_set(short);
extern void		 xerr(int, const char *, ...);
extern void		 xerrx(int, const char *, ...);
extern void		*xmalloc(size_t);
extern char		*xstrdup(char *);

int			 addtop10(char *, struct stats *);
int			 chooseruleset(void);
int			 choosewordfile(int);
int			 compar(const void *, const void *);
void			 freewords(void);
static struct hentry	*getentry(char *);
static int		 getfinfo(struct finfo *, int n, char *);
int			 hcompar(const void *, const void *);
void			 loadhighscores(char *);
static int		 parseline(char *, char **, char **);
void			 readconfig(void);
static void		 readfile(char *, int);
static void		 setoptions(char *, char *, int);

#ifdef TEST
char			*escstr(char *);
int			 loadwords(char *);
char			*unescstr(char *);
#else
static char		*escstr(char *);
static int		 loadwords(char *);
static char		*unescstr(char *);
#endif /* TEST */

extern int	 hfd;
extern char	*rankki[11];
extern char	*typorank[12];
extern char	*usedwordfile;
extern char	*word[WORD_MAX];
extern int	 wordcount;

char ruledir[MAXPATHLEN];
char worddir[MAXPATHLEN];

/*
 * Asks player about his/her name. The entered name must be
 * changed so tabs are escaped. During write process, the
 * high score file will be locked.
 *
 * Returns 0 on success.
 */
int
addtop10(char *wordlist, struct stats *stat)
{
	FILE *highf;
	int fd;
	char *rname, *uname;
	char name[21];

	/* Ask player until a name has been entered. */
	do {
		drawscreen();
		mvaddstr(3, 2, _("Enter your name:"));
		liima_mvgetnstr(3, 3 + strlen(_("Enter your name:")), name,
		    sizeof(name) - 1);
		name[sizeof(name) - 1] = '\0';
	} while (!strlen(name));

	rname = escstr(rules.fname);
	uname = escstr(name);

	flock(hfd, LOCK_EX);

	if ((fd = dup(hfd)) == -1)
		xerr(1, "addtop10: dup");

	if ((highf = fdopen(fd, "r+")) == NULL)
		xerr(1, "addtop10: fdopen");

	fseek(highf, 0, SEEK_END);
	fprintf(highf, "%lu\t%lu\t%lu\t%s\t%s\t%s\t%lu\t%u\n",
	    stat->score, stat->tcount, stat->wordswritten, uname, wordlist,
	    rname, (unsigned long)stat->duration, stat->sinit);

	if (fclose(highf) == EOF)
		xerr(1, "addtop10: fclose");

	flock(hfd, LOCK_UN);

	free(rname);
	free(uname);

	return 0;
}

/*
 * This function asks player about rule set of choice. All rule set
 * variables will be set.
 *
 * Returns ...
 * 0 on success
 * 1 if no rule set available
 * -1 if path to rule set is too long
 */
int
chooseruleset(void)
{
	int k, n;
	char fullpath[MAXPATHLEN];
	struct finfo namelist[WORDFILE_MAX];

	drawscreen();

	/* Default rule set must be selectable. */
	(void)strncpy(namelist[0].descr, _("default"),
	    sizeof(namelist[0].descr) - 1);
	namelist[0].descr[sizeof(namelist[0].descr) - 1] = '\0';
	(void)strncpy(namelist[0].name, "default",
	    sizeof(namelist[0].name) - 1);
	namelist[0].name[sizeof(namelist[0].name) - 1] = '\0';

	k = getfinfo(&(namelist[1]), WORDFILE_MAX - 1, "rule.");
	if (!k++) {
		mvaddstr(1, 5, _("No rule sets found..."));
		pressanykey(3, 5);
		return 1;
	}

	qsort((void *)(namelist + 1), k - 1, sizeof(namelist[0]), compar);

	if (!(k = fileselmenu(k, namelist, _("rule set")))) {
		defrule();
		return 0;
	}

	n = snprintf(fullpath, sizeof(fullpath), "%s/%s", ruledir,
	    namelist[k].name);
	if (n < 1 || (size_t)n > sizeof(fullpath) - 1)
		return -1;

	readfile(fullpath, 0);
	(void)strncpy(rules.fname, namelist[k].name, sizeof(rules.fname) - 1);
	rules.fname[sizeof(rules.fname) - 1] = '\0';
	(void)strncpy(rules.name, namelist[k].descr, sizeof(rules.name) - 1);
	rules.name[sizeof(rules.name) - 1] = '\0';

	return 0;
}

/*
 * A huge function that asks player about word list.
 * In this menu you can select found word lists for further
 * processing. The processing of the word list depends on
 * "operation":
 *
 * 0: loads words out of word list
 * 1: prints high scores for word list (and current rule set)
 *
 * Returns 0 on success or 1 on failure.
 */
int
choosewordfile(int operation)
{
	int a, k;
	const char *errstr;
	struct finfo namelist[WORDFILE_MAX];

	if (!(k = getfinfo(namelist, WORDFILE_MAX, "words."))) {
		drawscreen();
		mvaddstr(1, 5, _("No word lists found..."));
		pressanykey(3, 5);
		return 1;
		/* NOTREACHED */
	}

	qsort((void *)namelist, k, sizeof(namelist[0]), compar);

	k = fileselmenu(k, namelist, _("word list"));

	switch (operation) {
	case 1:
		loadhighscores(namelist[k].name);
		pressanykey(16, 3);
		return 0;
		/* NOTREACHED */
		break;
	case 0:
		a = loadwords(namelist[k].name);
		break;
	default:
		a = -1;
		break;
	}

	switch (a) {
	case 0: /* success */
		if (usedwordfile != NULL)
			free(usedwordfile);
		usedwordfile = xstrdup(namelist[k].name);
		return 0;
		/* NOTREACHED */
		break;
	case 1: /* snprintf */
		errstr = _("Path to file is too long.");
		break;
	case 2: /* fopen - errno */
		errstr = strerror(errno);
		break;
	case 3: /* loadwords - not enough words */
		errstr = _("Not enough words in word list for current settings.");
		break;
	case 5: /* resetscorefile - error during write */
		errstr = _("Cannot write to high score file.");
		break;
	default:
		errstr = _("Something went wrong. Contact your mother.");
		break;
	}

	xcolor_set(2);
	mvaddstr(3, 5, errstr);
	while (getch() == ERR)
		;

	return 1;
}

/* helper function for qsort */
int
compar(const void *p1, const void *p2)
{
	int retval;

	struct finfo *elem1, *elem2;

	elem1 = (struct finfo *)p1;
	elem2 = (struct finfo *)p2;

	if (!(retval = strcmp(elem1->descr, elem2->descr)))
		xerrx(1, "%s and %s have the same description!\n",
		    elem1->name, worddir, elem2->name);
	return retval;
}

/*
 * All entries in high score file must be valid. If a player
 * has created a word list with a file name that contains a tab
 * (or rule set with tab), the tab must be escaped.
 *
 * Returns converted string that must be passed to free() later on.
 */
#ifdef TEST
char *
#else
static char *
#endif /* TEST */
escstr(char *string)
{
	int pos;
	char *new, *p, *retval;

	if (string == NULL)
		return NULL;

	if ((new = calloc(2, strlen(string) + 1)) == NULL)
		xerr(1, "escstr: calloc");

	for (pos = 0, p = string; *p != '\0'; p++) {
		if (*p == '\\' || *p == '\t')
			new[pos++] = '\\';
		new[pos++] = *p;
	}
	new[pos] = '\0';

	retval = xstrdup(new);
	free(new);

	return retval;
}

/* Frees all allocated memory in word. */
void
freewords(void)
{
	while (wordcount) {
		free(word[--wordcount]);
		word[wordcount] = NULL;
	}
}

/*
 * This function converts string "line" into struct hentry. The converted
 * struct will be returned and must be passed to free() later on.
 *
 * Returns converted struct.
 */
static struct hentry *
getentry(char *line)
{
	int i, n, pos;
	char *p, *q, *t;
	struct hentry *entry;

	entry = xmalloc(sizeof(struct hentry));
	p = line;

	for (pos = 0; pos < 7; pos++) {
		/* do not allow empty field */
		if ((q = strchr(p, '\t')) == NULL || q == p)
			xerrx(1, "bad entry in high score detected");

		for (i = 1; (q - i) >= p && *(q - i) == '\\'; i++)
			;

		if (!(i % 2)) {
			p = q + 1;
			continue;
		}
		*q = '\0';

		switch(pos) {
		case 0: /* count */
			if (!(entry->count = cstrl(line)) && errno)
				xerrx(1, "count is invalid: %s\n", line);
			break;
		case 1: /* tcount */
			if (!(entry->tcount = cstrl(line)) && errno)
				xerrx(1, "tcount is invalid: %s\n", line);
			break;
		case 2: /* wordswritten */
			if (!(entry->wordswritten = cstrl(line)) && errno)
				xerrx(1, "wordswritten is invalid: %s\n", line);
			break;
		case 3: /* name */
			while (*line == ' ')
				line++;
			t = unescstr(line);
			n = snprintf(entry->name, sizeof(entry->name), "%s",
			    t);
			if (n < 1 || (size_t)n > sizeof(entry->name) - 1)
				xerrx(1, "name is invalid: %s\n", t);
			free(t);
			break;
		case 4: /* word list */
			while (*line == ' ')
				line++;
			t = unescstr(line);
			n = snprintf(entry->wordlist, sizeof(entry->wordlist),
			    "%s", t);
			if (n < 1 || (size_t)n > sizeof(entry->wordlist) - 1)
				xerrx(1, "word list is invalid: %s\n", t);
			free(t);
			break;
		case 5: /* mod */
			while (*line == ' ')
				line++;
			t = unescstr(line);
			n = snprintf(entry->mod, sizeof(entry->mod), "%s", t);
			if (n < 1 || (size_t)n > sizeof(entry->mod) - 1)
				xerrx(1, "game rule is invalid: %s\n", t);
			free(t);
			break;
		case 6: /* duration */
			if (!(entry->duration = cstrl(line)) && errno)
				xerrx(1, "duration is invalid: %s\n", line);
			break;
		default:
			xerrx(1, "too many fields in high score file");
			/* NOTREACHED */
			break;
		}
		line = p = q + 1;
	}
	/* sinit */
	if (!(entry->sinit = cstrl(line)) && errno)
		xerrx(1, "sinit is invalid: %s\n", line);

	return entry;
}

/*
 * getfinfo is used to gather specific file information. Possible
 * files are game rules and word lists - depending on "prefix".
 * namelist will be filled with all available options.
 *
 * Returns
 * -1 if wrong prefix has been supplied
 * number of game rules/word lists found
 */
static int
getfinfo(struct finfo *namelist, int n, char *prefix)
{
	DIR *dirp;
	FILE *fileinfo;
	int k, op, w;
	char *p, *path;
	char info[61], wordpath[MAXPATHLEN];
	struct dirent *dp;

	if (!strcmp(prefix, "rule.")) {
		op = 0;
		path = ruledir;
	} else if (!strcmp(prefix, "words.")) {
		op = 1;
		path = worddir;
	} else
		return -1;

	if ((dirp = opendir(path)) == NULL)
		xerr(1, "getfinfo: opendir: %s", path);

	k = 0;
	while ((dp = readdir(dirp)) != NULL && k < n) {
		if (strncmp(dp->d_name, prefix, strlen(prefix)))
			continue;

		w = snprintf(wordpath, sizeof(wordpath), "%s/%s", path,
		    dp->d_name);
		if (w < 1 || (size_t)w > sizeof(wordpath) - 1)
			continue;
		if ((fileinfo = fopen(wordpath, "r")) == NULL)
			continue;
		if (fgets(info, sizeof(info), fileinfo) != NULL) {
			if ((p = strchr(info, '\n')) != NULL)
				*p = '\0';
			if ((p = strchr(info, '\r')) != NULL)
				*p = '\0';
			p = info;
			if (!op) {
				if (info[0] != '#')
					xerrx(1, "invalid rule set");
				p++;
			}
			(void)strncpy(namelist[k].descr, p,
			    sizeof(namelist[k].descr) - 1);
			namelist[k].descr[sizeof(namelist[k].descr) - 1] =
			    '\0';
			(void)strncpy(namelist[k].name, dp->d_name,
			    sizeof(namelist[k].name) - 1);
			namelist[k].name[sizeof(namelist[k].name) - 1] = '\0';
			k++;
		}
		if (fclose(fileinfo) == EOF)
			xerr(1, "getfinfo: fclose: %s", wordpath);
	}
	(void)closedir(dirp);

	return k;
}

/* helper function for qsort */
int
hcompar(const void *entry1, const void *entry2)
{
	struct hentry **en1, **en2;
	unsigned long val1, val2;

	en1 = (struct hentry **)entry1;
	en2 = (struct hentry **)entry2;
	val1 = (*en1)->count;
	val2 = (*en2)->count;

	if (val1 < val2)
		return 1;
	if (val1 == val2)
		return 0;
	return -1;
}

/*
 * Loads high score for "sanafile", where "sanafile" contains the name
 * of the word list and gives out a high score list for current rule set.
 */
void
loadhighscores(char *sanafile)
{
	FILE *highfile;
	int fd, i, n, x;
	char line[2 * NAME_MAX + 100];
	char *p;
	struct hentry *entry;
	struct hentry **entries, **new;

	entries = NULL;

	flock(hfd, LOCK_EX);

	if ((fd = dup(hfd)) == -1)
		xerr(1, "loadhighscores: dup");
	if ((highfile = fdopen(fd, "r+")) == NULL)
		xerr(1, "loadhighscores: fdopen");
	rewind(highfile);

	n = 0;
	while (fgets(line, sizeof(line), highfile) != NULL) {
		if ((p = strchr(line, '\n')) == NULL)
			xerrx(1, "corrupt high score file");
		*p = '\0';
		if ((p = strchr(line, '\r')) != NULL)
			*p = '\0';
		entry = getentry(line);
		if (!strcmp(entry->wordlist, sanafile) && !strcmp(entry->mod,
		    rules.fname)) {
			if ((new = realloc(entries, (n + 1) *
			    sizeof(struct hentry *))) == NULL)
				xerr(1, NULL);
			entries = new;
			entries[n++] = entry;
		}
		else
			free(entry);
	}

	if (fclose(highfile) == EOF)
		xerr(1, "loadhighscores: fclose");

	flock(hfd, LOCK_UN);

	qsort(entries, n, sizeof(struct hentry *), hcompar);

	if ((new = realloc(entries, (n + 1) * sizeof(struct hentry *)))
	    == NULL)
		xerr(1, NULL);
	entries = new;
	entries[n] = NULL;

	drawscreen();
	xcolor_set(4);

	if (entries[0] == NULL) {
		mvaddstr(4, 3, _("No High Scores"));

		for (i = 0; i < n; i++)
			if (entries[i] != NULL) {
				free(entries[i]);
				entries[i] = NULL;
			}
		free(entries);
		return;
	}

	mvaddstr(1, 0, _(
"Top10  (score) (name)                   (level)   (cps)  (tcps) (typoinfos)"));
	xcolor_set(5);

	for (x = 0; entries[x] != NULL && x < 10; x++) {
		float ratio;
		int lvl, rank;

		if (!entries[x]->tcount)
			ratio = 0;
		else
			ratio = (1 - (float)(entries[x]->count +
			    entries[x]->wordswritten) / (entries[x]->tcount +
			    entries[x]->wordswritten)) * 100;

		xcolor_set(5);
		mvprintw(x + 2, 2, "%2d.", x + 1);
		mvprintw(x + 2, 7, "%6lu", entries[x]->count);
		mvaddstr(x + 2, 15, entries[x]->name);
		lvl = level(entries[x]->count);
		if (lvl >= 0 && lvl < 11)
			mvaddstr(x + 2, 40, rankki[lvl]);
		xcolor_set(5);

		if (entries[x]->duration) {
			mvprintw(x + 2, 50, "%2.3f",
			    (float)(entries[x]->count +
			    entries[x]->wordswritten) / entries[x]->duration
			    * 100.0f);
			mvprintw(x + 2, 57, "%2.3f",
			    (float)(entries[x]->tcount +
			    entries[x]->wordswritten) / entries[x]->duration
			    * 100.0f);
		} else {
			mvprintw(x + 2, 50, "%2.3f", 0.0f);
			mvprintw(x + 2, 57, "%2.3f", 0.0f);
		}

		mvprintw(x + 2, 64, "%2.2f%%", ratio);

		rank = typorankkaus(ratio);
		if (!entries[x]->count)
			rank = 0;
		mvaddstr(x + 2, 71, typorank[rank]);
	}

	for (i = 0; i < n; i++)
		if (entries[i] != NULL) {
			free(entries[i]);
			entries[i] = NULL;
		}
	free(entries);
}

/*
 * Reads in max WORD_MAX words out of file "filename" into word. If
 * less than 22 words are parsed, all words are freed and 1 is returned.
 * Words cannot be longer than 19 chars. No word duplicates allowed.
 *
 * First line of file "filename" will be ignored (description).
 *
 * XXX: loadwords is able to add words to an already populated
 * word array, I think this is not needed anywhere, so this
 * functionality will be removed.
 *
 * Returns ...
 * 0 on success.
 * 1 on snprintf failure (file path too long).
 * 2 on failure during file opening (errno set by fopen).
 * 3 if not enough words in wordfile.
 */
#ifdef TEST
int
#else
static int
#endif /* TEST */
loadwords(char *filename)
{
	FILE *wordlist;
	int ignore, j, n;
	char *myword[WORD_MAX], *p;
	char newword[19 + 3], wordpath[MAXPATHLEN];
	size_t i, mycount;

	n = snprintf(wordpath, sizeof(wordpath), "%s/%s", worddir, filename);
	if (n < 1 || (size_t)n > sizeof(wordpath) - 1)
		return 1;

	if ((wordlist = fopen(wordpath, "r")) == NULL)
		return 2;

	memset(myword, 0, sizeof(myword));
	mycount = 0;
	ignore = 1;
	while (fgets(newword, sizeof(newword), wordlist) != NULL) {
		/*
		 * We won't freak out about words which are against
		 * the rule (e.g. too long). Instead we will silently
		 * ignore them.
		 */
		if ((p = strchr(newword, '\n')) == NULL) {
			ignore = 1;
			continue;
		}
		if (ignore) {
			ignore = 0;
			continue;
		}

		*p = '\0';
		if ((p = strchr(newword, '\r')) != NULL)
			*p = '\0';

		if (strlen(newword) > 19 || !strlen(newword))
			continue;

		for (ignore = 0, p = newword; *p != '\0'; p++)
			if (!isprint(*p) || isspace(*p)) {
				ignore = 1;
				break;
			}
		if (ignore) {
			ignore = 0;
			continue;
		}

		if (mycount >= WORD_MAX)
			break;

		myword[mycount++] = xstrdup(newword);
	}
	if (fclose(wordlist) == EOF)
		xerr(1, "loadwords: fclose: %s", wordpath);

	for (i = 0; i < mycount; i++) {
		for (j = 0; j < WORD_MAX; j++) {
			if (word[j] == NULL) {
				if (wordcount < WORD_MAX) {
					wordcount++;
					word[j] = myword[i];
				}
				break;
			}
			if (!strcmp(myword[i], word[j])) {
				free(myword[i]);
				break;
			}
		}
		if (wordcount == WORD_MAX)
			break;
	}

	if (wordcount < 22) {
		/* not enough words */
		freewords();
		return 3;
	}
	return 0;
}

/*
 * Used to fill in option and value out of line. "line" should be a line
 * out of configuration file or rule set file.
 *
 * IMPORTANT: "line" _will_ be changed, as option and value only point
 * at the correct position of line.
 *
 * Returns 0 on success or 1 on failure.
 */
static int
parseline(char *line, char **option, char **value)
{
	char *p, *v;

	/* By default, no option and no value will be returned. */
	*option = *value = NULL;

	/*
	 * First off, the line will be "normalized", i.e. all
	 * unneeded characters (white spaces in front of option,
	 * \n and possible \r and comments [beginning with #]).
	 * If there is no \n, the line was not fully red, so
	 * let's drop it completely.
	 */
	if ((p = strchr(line, '\n')) == NULL)
		return 1;
	*p = '\0';
	if ((p = strchr(line, '\r')) != NULL)
		*p = '\0';
	if ((p = strchr(line, '#')) != NULL)
		*p = '\0';
	while (*line != '\0' && isblank(*line))
		line++;

	/* Anything left? */
	if (!strlen(line))
		return 0;

	/*
	 * Locate the first "=" and use p for the oPtion
	 * and v for the Value.
	 */
	if ((v = strchr(line, '=')) == NULL || v == line)
		return 1;

	*v = '\0';
	p = v - 1;
	v++;

	/* Remove all trailing whitespaces between option and equal sign. */
	while (p != line && isblank(*p))
		*(p--) = '\0';

	/* Remove all whitespaces between equal sign and value. */
	while (*v != '\0' && isblank(*v))
		v++;

	/* Remove all trailing whitespaces. */
	if ((p = strchr(v, '\0')) == NULL || p == v)
		return 1;
	p--;
	while (p != v && isblank(*p))
		*(p--) = '\0';

	*option = line;
	*value = v;

	return 0;
}

/*
 * Opens system wide and user specific configuration file. If there is
 * a user configuration file, a user specific high score file will be
 * opened, too. This way, the system wide high score won't be tampered with.
 */
void
readconfig(void)
{
	int n;
	char *envhome;
	char userhigh[MAXPATHLEN], userconf[MAXPATHLEN];
	struct stat sb;

	n = snprintf(ruledir, sizeof(ruledir), "%s", RULEDIR);
	if (n < 0 || (size_t)n > sizeof(ruledir) - 1) {
		ruledir[0] = '.';
		ruledir[1] = '\0';
	}
	n = snprintf(worddir, sizeof(worddir), "%s", WORDDIR);
	if (n < 0 || (size_t)n > sizeof(worddir) - 1) {
		worddir[0] = '.';
		worddir[1] = '\0';
	}

	readfile(CONFIGFILE, 1);

	if ((envhome = getenv("HOME")) == NULL)
		return;

	n = snprintf(userconf, sizeof(userconf), "%s/.typespeed/config",
	    envhome);
	if (n < 1 || (size_t)n > sizeof(userconf) - 1)
		return;

	if (stat(userconf, &sb) || (sb.st_mode & S_IFMT) != S_IFREG)
		return;

	n = snprintf(userhigh, sizeof(userhigh), "%s/.typespeed/score",
	    envhome);
	if (n < 1 || (size_t)n > sizeof(userhigh) - 1)
		return;

	/*
	 * Open a user writable high score.
	 * This can be a symbolic link to system-wide high score
	 * file. Protect system-wide high score file with group
	 * write permissions: privileged gid already dropped.
	 */
	if (close(hfd) == -1)
		xerr(1, "readconfig: close");
	if ((hfd = open(userhigh, O_RDWR, 0)) == -1)
		xerr(1, "readconfig: open: %s", userhigh);

	readfile(userconf, 1);
}

/*
 * Function used to open configuration and game rule files and to
 * set options with function setoptions.
 */
static void
readfile(char *filename, int op)
{
	FILE *file;
	unsigned long lineno;
	char *option, *value;
	char line[10 + MAXPATHLEN + 3]; /* 10 = strlen("worddir = ") */

	lineno = 0;

	if ((file = fopen(filename, "r")) == NULL)
		return;

	while (fgets(line, sizeof(line), file) != NULL) {
		lineno++;
		if (parseline(line, &option, &value))
			xerrx(1, "Error in %s line %lu\n", filename, lineno);
		if (option != NULL && value != NULL)
			setoptions(option, value, op);
	}

	if (fclose(file) == EOF)
		xerr(1, "readfile: fclose: %s", filename);
}

/*
 * setoptions sets actually configuration and rule sets. If a
 * configuration or rule set has been opened decides "op":
 * 0: game rule
 * 1: configuration file
 */
static void
setoptions(char *option, char *value, int op)
{
	unsigned long val;

	if (op) { /* configuration file */
		if (!strcmp(option, "cheat")) {
			if (!strcmp(value, "yes"))
				opt.cheat = 1;
			else
				opt.cheat = 0;
		}
		else if (!strcmp(option, "ruledir")) {
			(void)strncpy(ruledir, value, sizeof(ruledir) - 1);
			ruledir[sizeof(ruledir) - 1] = '\0';
		}
		else if (!strcmp(option, "worddir")) {
			(void)strncpy(worddir, value, sizeof(worddir) - 1);
			worddir[sizeof(worddir) - 1] = '\0';
		}
	}
	else { /* game rule */
		if (!strcmp(option, "misses")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "misses invalid: %s\n", value);
			if (val > 0 && val < 100)
				rules.misses = val;
		}
		else if (!strcmp(option, "min word length")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "min word length invalid: %s\n",
				    value);
			if (val > 0 && val < 20)
				rules.minlen = val;
		}
		else if (!strcmp(option, "max word length")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "max word length invalid: %s\n",
				    value);
			if (val > 0 && val < 20)
				rules.maxlen = val;
		}
		else if (!strcmp(option, "min words")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "min words invalid: %s\n", value);
			if (val > 0 && val < 23)
				rules.minwords = val;
		}
		else if (!strcmp(option, "max words")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "max words invalid: %s\n", value);
			if (val > 0 && val < 23)
				rules.maxwords = val;
		}
		else if (!strcmp(option, "highscore")) {
			if (!strcmp(value, "yes"))
				rules.hightype = 1;
			else
				rules.hightype = 0;
		}
		else if (!strcmp(option, "min score")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "min score invalid: %s\n", value);
			rules.minscore = val;
		}
		else if (!strcmp(option, "min speed")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "min speed invalid: %s\n", value);
			if (val < 100)
				rules.minspeed = val;
		}
		else if (!strcmp(option, "max speed")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "max speed invalid: %s\n", value);
			if (val < 100)
				rules.maxspeed = val;
		}
		else if (!strcmp(option, "name")) {
			(void)strncpy(rules.name, value,
			    sizeof(rules.name) - 1);
			rules.name[sizeof(rules.name) - 1] = '\0';
		}
		else if (!strcmp(option, "step")) {
			if (!(val = cstrl(value)) && errno)
				xerrx(1, "step invalid: %s\n", value);
			if (val < 1000)
				rules.step = val;
		}
		else if (!strcmp(option, "smooth")) {
			if (!strcmp(value, "yes"))
				rules.smooth = 1;
			else
				rules.smooth = 0;
		}
	}
}

/*
 * Removes escaped tabs and returns converted string.
 * This string must be passed to free later on!
 */
#ifdef TEST
char *
#else
static char *
#endif /* TEST */
unescstr(char *string)
{
	int pos;
	char *new, *p, *retval;

	if (string == NULL)
		return NULL;

	new = xmalloc(strlen(string) + 1);

	for (pos = 0, p = string; *p != '\0'; p++, pos++) {
		if (*p == '\\') {
			if (*(p + 1) == '\0')
				xerrx(1, "error in unescstr");
			else
				p++;
		}
		new[pos] = *p;
	}
	new[pos] = '\0';

	retval = xstrdup(new);
	free(new);

	return retval;
}

