/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006-2007   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * network.c - network for typespeed
 *
 * thanks goto Steinar H. Gunderson <sgunderson@bigfoot.com>
 * for the basic functions which we got from him
 */

#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_SYS_IOCTL_H
	#include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */

#ifdef HAVE_SYS_TYPES_H
	#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */

#ifdef HAVE_SYS_SOCKET_H
	#include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */

#ifdef HAVE_NETINET_IN_H
	#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */

#ifdef HAVE_ARPA_INET_H
	#include <arpa/inet.h>
#endif /* HAVE_ARPA_INET_H */

#include <curses.h>
#include <errno.h>

#ifdef HAVE_LIMITS_H
	#include <limits.h>
#endif /* HAVE_LIMITS_H */

#ifdef HAVE_NETDB_H
	#include <netdb.h>
#endif /* HAVE_NETDB_H */

#include <stdio.h>

#ifdef HAVE_STDLIB_H
	#include <stdlib.h>
#endif /* HAVE_STDLIB_H */

#ifdef HAVE_STRING_H
	#include <string.h>
#endif /* HAVE_STRING_H */

#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include "gettext.h"
#include "typespeed.h"

#define _(string)	gettext(string)

extern int	addnewword(char *);
extern void	endcursestuff(void);
extern int	r(int);
extern void	xerr(int, const char *, ...);
extern void	xerrx(int, const char *, ...);

void		closenetwork(void);
int		initnetwork(char *, int);
static int	netident(int, int, int);
int		netrecvword(int);
void		netsendword(int, char *);
void		netswapscore(int, struct stats *, struct stats *);
void		netwaitgame(int);

static int ttss;

/*
 * Closes network socket and sets global ttss = 0.
 */
void
closenetwork(void)
{
	close(ttss);
	ttss = 0;
}

/*
 * In this function, client and server are set up. Output must be possible
 * on command line and curses mode, because command line options can be
 * passed to typespeed - this function would be called before cursesinit.
 * If opt.client is set, client specific code will be executed. In this case,
 * serv will be contacted.
 * "graph" tells initnetwork if graphic mode has been initialized (1) or
 * not (0).
 *
 * Returns network socket or -69 on error.
 */
int
initnetwork(char *serv, int graph)
{
	int i, one, sock, ss;
	struct sockaddr_in addr;

	one = 1;

	if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
		xerr(1, "initnetwork: socket");

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = PF_INET;
	addr.sin_port = htons(opt.port);

	if (opt.client) {
		/* client code */
		if (!inet_aton(serv, &addr.sin_addr)) {
			struct hostent *he;

			if (graph) {
				mvprintw(10, 30, _("Looking up %s..."), serv);
				refresh();
			} else
				printf(_("Looking up %s...\n"), serv);

			if ((he = gethostbyname(serv)) == NULL) {
				endcursestuff();
				fprintf(stderr,
				    "initnetwork: gethostbyname: %s: ", serv);
				herror(NULL);
				exit(1);
			}
			memcpy(&addr.sin_addr, he->h_addr, he->h_length);
		}

		if (graph) {
			mvaddstr(10,30, _("Connecting to typespeed server at "));
			mvprintw(11,30, "%s (port %d)...",
			    inet_ntoa(addr.sin_addr), opt.port);
			refresh();
		} else {
			fputs(_("Connecting to typespeed server at "), stdout);
			printf("%s (port %d)...\n", inet_ntoa(addr.sin_addr),
			    opt.port);
		}

		if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
			xerr(1, "initnetwork: connect: %s", serv);

		if ((i = netident(sock, 0, graph)))
			return i;
	} else {
		/* server code */
		addr.sin_addr.s_addr = INADDR_ANY;

		if ((ss = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
			xerr(1, "initnetwork: socket");

		setsockopt(ss, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
		if (bind(ss, (struct sockaddr *)&addr,
		    sizeof(struct sockaddr)) == -1)
			xerr(1, "initnetwork: bind");

		if (listen(ss, 1) == -1)
			xerr(1, "initnetwork: listen");

		if (graph) {
			mvaddstr(10, 30, _("Waiting for connections..."));
			refresh();
		} else
			puts(_("Waiting for connections..."));

		if ((sock = accept(ss, INADDR_ANY, NULL)) == -1)
			xerr(1, "initnetwork: accept");

		ttss = ss;

		if ((i = netident(sock, ss, graph))) {
			close(ss);
			return i;
		}
	}

	ioctl(sock, FIONBIO, &one);

	return sock;
}

/*
 * Sends and receives typespeed identification string.
 *
 * Returns 0 on success (remote uses typespeed and connection successful)
 * or -69 on error.
 */
static int
netident(int typesock, int tss, int graph)
{
	int i, retval;
	char buf[BUFSIZ];

	i = snprintf(buf, sizeof(buf), "TYPESPEED %s\n", TVERSION);
	if (i < 10 || (size_t)i > sizeof(buf) - 1)
		xerrx(1, "netident: malformed version");

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1)
		xerr(1, "netident: send");
	if ((size_t)i != strlen(buf))
		xerrx(1, "%d out of %zd bytes sent", i, strlen(buf));

	memset(buf, 0, sizeof(buf));

	if (graph) {
		mvaddstr(10, 30, _("Waiting for other party to identify..."));
		refresh();
	} else {
		puts(_("Waiting for other party to identify..."));
	}

	i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
	buf[sizeof(buf) - 1] = '\0';

	if (i == -1 && errno != EWOULDBLOCK)
		xerr(1, "netident: recv");

	retval = -69;

	if (i > 0) {
		recv(typesock, buf, strlen(buf) + 1, 0);
		if (!strncmp(buf, "TYPESPEED ", strlen("TYPESPEED "))) {
			if (graph) {
				mvprintw(10, 30, _("Remote is using: %s\n"), buf);
				refresh();
			} else
				printf(_("Remote is using: %s\n"), buf);
			retval = 0;
		} else {
			if (graph) {
				mvaddstr(10, 30, _("Remote is not TYPESPEED"));
				refresh();
			} else
				puts(_("Remote is not TYPESPEED"));
			close(tss);
		}
	}

	return retval;
}

/*
 * Interprets incoming words.
 * If incoming word is anything but " ENDOFTHEGAME", trailing newline
 * will be removed and word passed to addnewword.
 *
 * Returns 0 on success (word added) or 2 if network should be shut down.
 */
int
netrecvword(int typesock)
{
	int retval;
	ssize_t i;
	char *j;
	char buf[BUFSIZ];

	retval = 0;

	i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
	buf[sizeof(buf) - 1] = '\0';

	if (i == -1 && errno != EWOULDBLOCK)
		xerr(1, "netrecvword: recv");
	if ((j = strchr(buf, '\n')) != NULL)
		*(j + 1) = '\0';
	if (i > 0) {
		recv(typesock, buf, strlen(buf) + 1, 0);
		if (!strncmp(buf, " ENDOFTHEGAME", strlen(" ENDOFTHEGAME")))
			retval = 2;
		else {
			if ((j = strchr(buf, '\n')) != NULL)
				*j = '\0';
			addnewword(buf);
		}
	}

	return retval;
}

/*
 * Sends word "wordstring" through socket "typesock" to opponent.
 */
void
netsendword(int typesock, char *wordstring)
{
	int i;
	char buf[BUFSIZ];

	i = snprintf(buf, sizeof(buf), "%s\n", wordstring);
	if (i < 1 || (size_t)i > sizeof(buf) - 1)
		xerrx(1, "netsendword: malformed word %s", wordstring);

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1)
		xerr(1, "netsendword: send");
	if ((size_t)i != strlen(buf))
		xerrx(1, "%d out of %zd bytes sent", i, strlen(buf));
}

/*
 * Converts current stat into network protocol text (a.k.a. simple string)
 * and sends to opponent. Next to this, opponent's score will be received.
 *
 * stat2 will be filled with opponents score.
 */
void
netswapscore(int typesock, struct stats *stat, struct stats *stat2)
{
	int i;
	char *number, *p;
	char buf[BUFSIZ];

	i = snprintf(buf, sizeof(buf), "SCORE: %lu %f %f %f\n", stat->score,
	    stat->speed, stat->totalspeed, stat->ratio);
	if (i < 11 || (size_t)i > sizeof(buf) - 1)
		xerrx(1, "netswapscore: malformed line");

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1)
		xerr(1, "netswapscore: send");
	if ((size_t)i != strlen(buf))
		xerrx(1, "%d out of %zd bytes sent", i, strlen(buf));
	memset(buf, 0, sizeof(buf));

	for (;;) {
		i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
		buf[sizeof(buf) - 1] = '\0';

		if (i == -1 && errno != EWOULDBLOCK)
			xerr(1, "netswapscore: recv");
		if (i > 0) {
			recv(typesock, buf, strlen(buf) + 1, 0);
		        if (!strncmp(buf, "SCORE: ", strlen("SCORE: ")))
				break;
		}
	}

	memset(stat2, 0, sizeof(stat2));

	number = p = buf + strlen("SCORE: ");
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->score = strtol(number, NULL, 10);

	number = ++p;
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->speed = (float)atof(number);

	number = ++p;
	if ((p = strchr(p, ' ')) == NULL)
		return;
	*p = '\0';
	stat2->totalspeed = (float)atof(number);

	number = ++p;
	stat2->ratio = (float)atof(number);
}

/*
 * Sends typespeed with version to opponent. This function is used to
 * send opponent that something has happend (word list selected).
 */
void
netwaitgame(int typesock)
{
	int i;
	char buf[BUFSIZ];

	i = snprintf(buf, sizeof(buf), "TYPESPEED %s\n", TVERSION);
	if (i < 10 || (size_t)i > sizeof(buf) - 1)
		xerrx(1, "netwaitgame: malformed line");

	if ((i = send(typesock, buf, strlen(buf), 0)) == -1)
		xerr(1, "netwaitgame: send");
	if ((size_t)i != strlen(buf))
		xerrx(1, "%d out of %zd bytes sent", i, strlen(buf));

	for (;;) {
		i = recv(typesock, buf, sizeof(buf), MSG_PEEK);
		buf[sizeof(buf) - 1] = '\0';

		if (i == -1 && errno != EWOULDBLOCK)
			xerr(1, "netwaitgame: recv");
		if (i > 0) {
			recv(typesock, buf, strlen(buf) + 1, 0);
                	if (!strncmp(buf, "TYPESPEED", strlen("TYPESPEED")))
				break;
		}
	}
}

